"""Test runner with colored output and reporting."""

from __future__ import annotations

import sys
import time
from typing import TYPE_CHECKING, TextIO

from rich.console import Console
from rich.table import Table
from typing_extensions import Self

from webdav_tck.framework.result import TestResultInfo, WebdavTckTestResult
from webdav_tck.framework.suite import (
    WebdavTckContext,
    WebdavTckTestCase,
    WebdavTckTestSuite,
)

if TYPE_CHECKING:
    from pathlib import Path


class TestRunner:
    """Runs test suites and reports results.

    Provides colored terminal output and debug logging.
    """

    def __init__(
        self,
        quiet: bool = False,
        use_color: bool | None = None,
        debug_log: Path | None = None,
    ):
        """Initialize test runner.

        Args:
            quiet: Use abbreviated output
            use_color: Force color output (None = auto-detect)
            debug_log: Path to debug log file
        """
        self.quiet = quiet
        self.console = Console(force_terminal=use_color, file=sys.stdout)
        self.debug_log = debug_log
        self.debug_file: TextIO | None = None

        # Result tracking
        self.results: list[TestResultInfo] = []
        self.current_suite: str | None = None

    def __enter__(self) -> Self:
        """Open debug log file if specified."""
        if self.debug_log:
            self.debug_file = open(self.debug_log, "w", encoding="utf-8")
        return self

    def __exit__(self, *args: object) -> None:
        """Close debug log file."""
        if self.debug_file:
            self.debug_file.close()

    def log_debug(self, message: str) -> None:
        """Write message to debug log."""
        if self.debug_file:
            timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
            self.debug_file.write(f"[{timestamp}] {message}\n")
            self.debug_file.flush()

    async def run_suite(self, suite: WebdavTckTestSuite) -> list[TestResultInfo]:
        """Run a test suite and return results.

        Args:
            suite: The test suite to run

        Returns:
            List of test results
        """
        self.current_suite = suite.name
        suite_results: list[TestResultInfo] = []

        if not self.quiet:
            self.console.print(f"\n[bold cyan]Running {suite.name}[/bold cyan]")

        # Run setup
        try:
            suite.run_setup()
        except Exception as e:
            self.console.print(f"[bold red]Setup failed: {e}[/bold red]")
            return suite_results

        # Run tests
        context = WebdavTckContext()
        stop_suite = False

        for test in suite.tests:
            if stop_suite:
                result_info = TestResultInfo(
                    test.name, WebdavTckTestResult.SKIP, "Suite stopped"
                )
                suite_results.append(result_info)
                continue

            context.clear()
            result = await self._run_test(test, suite, context)

            result_info = TestResultInfo(
                test.name, result, context.error_message, context.warnings
            )
            suite_results.append(result_info)
            self.results.append(result_info)

            self._print_result(test.name, result_info)

            if result.should_stop_suite():
                stop_suite = True

        # Run teardown
        try:
            suite.run_teardown()
        except Exception as e:
            self.console.print(f"[bold red]Teardown failed: {e}[/bold red]")

        return suite_results

    async def _run_test(
        self,
        test: WebdavTckTestCase,
        suite: WebdavTckTestSuite,
        context: WebdavTckContext,
    ) -> WebdavTckTestResult:
        """Run a single test.

        Args:
            test: The test to run
            suite: The parent suite
            context: Test context for warnings/errors

        Returns:
            Test result
        """
        self.log_debug(f"Running test: {suite.name}.{test.name}")

        try:
            # Call test function with session if available
            if suite.session:
                result = await test.func(suite.session, context)
            else:
                result = await test.func(context)

            # Convert int to WebdavTckTestResult
            if isinstance(result, int):
                result = WebdavTckTestResult(result)

            self.log_debug(f"Test {test.name} result: {result.name}")
            return result

        except Exception as e:
            self.log_debug(f"Test {test.name} exception: {e}")
            context.context(f"Exception: {e}")
            return WebdavTckTestResult.FAIL

    def _print_result(self, name: str, result: TestResultInfo) -> None:
        """Print test result to console.

        Args:
            name: Test name
            result: Test result info
        """
        if self.quiet:
            # Abbreviated output
            symbol = self._get_result_symbol(result.result)
            self.console.print(symbol, end="")
            return

        # Full output
        status_text = self._get_result_text(result.result)
        self.console.print(f"  {name}: {status_text}")

        # Print warnings
        for warning in result.warnings:
            self.console.print(f"    [yellow]Warning: {warning}[/yellow]")

        # Print error message
        if result.message:
            self.console.print(f"    [red]{result.message}[/red]")

    def _get_result_symbol(self, result: WebdavTckTestResult) -> str:
        """Get single-character symbol for result."""
        symbols = {
            WebdavTckTestResult.OK: "[green].[/green]",
            WebdavTckTestResult.FAIL: "[red]F[/red]",
            WebdavTckTestResult.FAILHARD: "[red]![/red]",
            WebdavTckTestResult.SKIP: "[yellow]s[/yellow]",
            WebdavTckTestResult.SKIPREST: "[yellow]S[/yellow]",
        }
        return symbols.get(result, "?")

    def _get_result_text(self, result: WebdavTckTestResult) -> str:
        """Get colored text for result."""
        texts = {
            WebdavTckTestResult.OK: "[green]PASS[/green]",
            WebdavTckTestResult.FAIL: "[red]FAIL[/red]",
            WebdavTckTestResult.FAILHARD: "[red]FAIL (hard)[/red]",
            WebdavTckTestResult.SKIP: "[yellow]SKIP[/yellow]",
            WebdavTckTestResult.SKIPREST: "[yellow]SKIP (rest)[/yellow]",
        }
        return texts.get(result, "[dim]UNKNOWN[/dim]")

    def print_summary(self) -> None:
        """Print summary of all test results."""
        if not self.results:
            return

        if self.quiet:
            self.console.print()  # Newline after dots

        # Count results
        total = len(self.results)
        passed = sum(1 for r in self.results if r.result == WebdavTckTestResult.OK)
        failed = sum(1 for r in self.results if r.result.is_failure())
        skipped = sum(1 for r in self.results if r.result.is_skip())

        # Create summary table
        table = Table(title="\nTest Summary", show_header=True)
        table.add_column("Status", style="bold")
        table.add_column("Count", justify="right")
        table.add_column("Percentage", justify="right")

        table.add_row("Total", str(total), "100%")
        table.add_row(
            "Passed",
            str(passed),
            f"{passed * 100 // total}%" if total > 0 else "0%",
            style="green",
        )

        if failed > 0:
            table.add_row(
                "Failed",
                str(failed),
                f"{failed * 100 // total}%" if total > 0 else "0%",
                style="red",
            )

        if skipped > 0:
            table.add_row(
                "Skipped",
                str(skipped),
                f"{skipped * 100 // total}%" if total > 0 else "0%",
                style="yellow",
            )

        self.console.print(table)

        # Print failed tests details
        failed_tests = [r for r in self.results if r.result.is_failure()]
        if failed_tests and not self.quiet:
            self.console.print("\n[bold red]Failed Tests:[/bold red]")
            for result in failed_tests:
                self.console.print(f"  • {result.name}")
                if result.message:
                    self.console.print(f"    {result.message}")

    def get_exit_code(self) -> int:
        """Get exit code based on test results.

        Returns:
            0 if all tests passed, 1 otherwise
        """
        return 1 if any(r.result.is_failure() for r in self.results) else 0



"""webdav_tck: WebDAV server protocol compliance test suite."""

from __future__ import annotations

try:
    from webdav_tck._version import version as __version__
except ImportError:
    __version__ = "0.0.0+unknown"

__all__ = ["__version__"]



"""CLI entry point for litmus test suite."""

from __future__ import annotations

import asyncio
import sys
from pathlib import Path

import click

from webdav_tck.framework import TestRunner
from webdav_tck.session import WebdavTckSession
from webdav_tck.suites.basic import create_basic_suite
from webdav_tck.suites.copymove import create_copymove_suite
from webdav_tck.suites.locks import create_locks_suite
from webdav_tck.suites.props import create_props_suite


async def run_litmus(
    url: str,
    username: str | None,
    password: str | None,
    proxy: str | None,
    system_proxy: bool,
    client_cert: str | None,
    insecure: bool,
    quiet: bool,
    colour: bool | None,
    suites: str | None,
) -> int:
    """Run litmus test suites.

    Args:
        url: WebDAV server URL
        username: Optional username
        password: Optional password
        proxy: Optional proxy URL
        system_proxy: Use system proxy
        client_cert: Optional client certificate
        insecure: Ignore SSL verification
        quiet: Abbreviated output
        colour: Force color output (None = auto)
        suites: Comma-separated list of suites

    Returns:
        Exit code (0 = success, 1 = failures)
    """
    # Determine which suites to run
    available_suites = ["basic", "copymove", "props", "locks"]
    if suites:
        suite_list = [s.strip() for s in suites.split(",")]
        suite_list = [s for s in suite_list if s in available_suites]
    else:
        suite_list = available_suites

    if not suite_list:
        click.echo("No valid test suites specified", err=True)
        return 1

    # Set up debug log
    debug_log = Path("debug.log")

    # Create session
    session = WebdavTckSession(
        url=url,
        username=username,
        password=password,
        verify_ssl=not insecure,
        proxy=proxy if not system_proxy else None,
        client_cert=client_cert,
        debug_log=debug_log,
    )

    # Create test runner
    with TestRunner(quiet=quiet, use_color=colour, debug_log=debug_log) as runner:
        try:
            # Initialize session
            async with session:
                click.echo(f"Testing WebDAV server: {url}")
                if username:
                    click.echo(f"Using authentication: {username}")

                # Discover capabilities
                runner.log_debug(f"Connecting to {url}")
                await session.discover_capabilities()

                if session.class2:
                    click.echo("Server supports WebDAV Class 2 (locking)")
                else:
                    click.echo("Server supports WebDAV Class 1 (basic)")

                # Create test collection
                await session.create_test_collection()
                runner.log_debug(f"Created test collection: {session.base_path}")

                # Run test suites
                for suite_name in suite_list:
                    if suite_name == "basic":
                        suite = create_basic_suite(session)
                        await runner.run_suite(suite)
                    elif suite_name == "copymove":
                        suite = create_copymove_suite(session)
                        await runner.run_suite(suite)
                    elif suite_name == "props":
                        suite = create_props_suite(session)
                        await runner.run_suite(suite)
                    elif suite_name == "locks":
                        suite = create_locks_suite(session)
                        await runner.run_suite(suite)

                # Cleanup
                await session.cleanup_test_collection()

            # Print summary
            runner.print_summary()

            return runner.get_exit_code()

        except KeyboardInterrupt:
            click.echo("\n\nInterrupted by user", err=True)
            return 130
        except Exception as e:
            click.echo(f"\n\nFatal error: {e}", err=True)
            runner.log_debug(f"Fatal error: {e}")
            return 1


@click.command()
@click.argument("url")
@click.argument("username", required=False)
@click.argument("password", required=False)
@click.option("--proxy", "-p", help="Proxy server URL")
@click.option(
    "--system-proxy", "-s", is_flag=True, help="Use system proxy configuration"
)
@click.option("--client-cert", "-c", help="PKCS#12 client certificate path")
@click.option(
    "--insecure", "-i", is_flag=True, help="Ignore TLS certificate verification"
)
@click.option("--quiet", "-q", is_flag=True, help="Use abbreviated output")
@click.option(
    "--colour/--no-colour",
    default=None,
    help="Force color output on/off (default: auto-detect)",
)
@click.option("--suites", help="Comma-separated list of test suites to run")
@click.version_option()
def main(
    url: str,
    username: str | None,
    password: str | None,
    proxy: str | None,
    system_proxy: bool,
    client_cert: str | None,
    insecure: bool,
    quiet: bool,
    colour: bool | None,
    suites: str | None,
) -> None:
    """Run litmus WebDAV test suite against URL.

    \b
    Examples:
        litmus http://localhost/webdav/
        litmus http://localhost/webdav/ user password
        litmus --insecure https://localhost/webdav/
        litmus --quiet --suites=basic http://localhost/webdav/

    The test suite will create a collection called 'litmus' at the
    specified URL and run various WebDAV protocol compliance tests.
    """
    # Validate URL
    if not url.startswith("http://") and not url.startswith("https://"):
        click.echo("Error: URL must start with http:// or https://", err=True)
        sys.exit(1)

    # Validate credentials
    if username and not password:
        click.echo("Error: Password required when username is provided", err=True)
        sys.exit(1)

    if password and not username:
        click.echo("Error: Username required when password is provided", err=True)
        sys.exit(1)

    # Run async main
    exit_code = asyncio.run(
        run_litmus(
            url=url,
            username=username,
            password=password,
            proxy=proxy,
            system_proxy=system_proxy,
            client_cert=client_cert,
            insecure=insecure,
            quiet=quiet,
            colour=colour,
            suites=suites,
        )
    )

    sys.exit(exit_code)


if __name__ == "__main__":
    main()



"""Test suite and test case classes."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from webdav_tck.framework.result import WebdavTckTestResult


class WebdavTckContext:
    """Context for a test execution.

    Collects warnings and error messages during test execution.
    """

    def __init__(self) -> None:
        self.warnings: list[str] = []
        self.error_message: str | None = None

    def warning(self, message: str) -> None:
        """Add a warning message."""
        self.warnings.append(message)

    def context(self, message: str) -> None:
        """Set error context message."""
        self.error_message = message

    def clear(self) -> None:
        """Clear all messages."""
        self.warnings.clear()
        self.error_message = None


class WebdavTckTestCase:
    """A single test case.

    Wraps a test function with metadata.
    """

    def __init__(
        self,
        func: Callable[..., Awaitable[WebdavTckTestResult] | WebdavTckTestResult | int],
        name: str | None = None,
        check_leaks: bool = False,
    ):
        self.func = func
        self.name = name or func.__name__
        self.check_leaks = check_leaks
        self._doc = func.__doc__

    def __repr__(self) -> str:
        return f"TestCase({self.name!r})"

    @property
    def description(self) -> str | None:
        """Get test description from docstring."""
        return self._doc.strip() if self._doc else None


class WebdavTckTestSuite:
    """A collection of related test cases.

    Maps to a test suite module in the original C implementation
    (e.g., basic, copymove, props, locks).
    """

    def __init__(self, name: str, session: Any = None):
        self.name = name
        self.session = session
        self.tests: list[WebdavTckTestCase] = []
        self._setup: Callable[[], None] | None = None
        self._teardown: Callable[[], None] | None = None

    def add_test(
        self,
        func: Callable[..., WebdavTckTestResult | int],
        name: str | None = None,
        check_leaks: bool = False,
    ) -> None:
        """Add a test to this suite."""
        self.tests.append(WebdavTckTestCase(func, name, check_leaks))

    def setup(self, func: Callable[[], None]) -> Callable[[], None]:
        """Decorator to register a setup function."""
        self._setup = func
        return func

    def teardown(self, func: Callable[[], None]) -> Callable[[], None]:
        """Decorator to register a teardown function."""
        self._teardown = func
        return func

    def run_setup(self) -> None:
        """Run setup function if defined."""
        if self._setup:
            self._setup()

    def run_teardown(self) -> None:
        """Run teardown function if defined."""
        if self._teardown:
            self._teardown()

    def __repr__(self) -> str:
        return f"TestSuite({self.name!r}, {len(self.tests)} tests)"



"""Test result types and classes."""

from __future__ import annotations

from enum import IntEnum


class WebdavTckTestResult(IntEnum):
    """Test result status codes.

    Maps to the original C litmus result codes:
    - OK: Test passed
    - FAIL: Test failed
    - FAILHARD: Test failed, skip remaining tests in suite
    - SKIP: Test skipped (precondition not met)
    - SKIPREST: Skipped, and skip all remaining tests in suite
    """

    OK = 0
    FAIL = 1
    FAILHARD = 2
    SKIP = 3
    SKIPREST = 4

    def is_failure(self) -> bool:
        """Return True if this result represents a failure."""
        return self in (WebdavTckTestResult.FAIL, WebdavTckTestResult.FAILHARD)

    def is_skip(self) -> bool:
        """Return True if this result represents a skip."""
        return self in (WebdavTckTestResult.SKIP, WebdavTckTestResult.SKIPREST)

    def should_stop_suite(self) -> bool:
        """Return True if this result should stop the test suite."""
        return self in (WebdavTckTestResult.FAILHARD, WebdavTckTestResult.SKIPREST)


class TestResultInfo:
    """Information about a test result."""

    def __init__(
        self,
        name: str,
        result: WebdavTckTestResult,
        message: str | None = None,
        warnings: list[str] | None = None,
    ):
        self.name = name
        self.result = result
        self.message = message
        self.warnings = warnings or []

    def __repr__(self) -> str:
        return f"TestResultInfo(name={self.name!r}, result={self.result.name})"



"""Properties test suite.

Ports src/props.c from the original litmus C implementation.
Tests property manipulation via PROPFIND and PROPPATCH operations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from webdav_tck.framework import (
    WebdavTckContext,
    WebdavTckTestResult,
    WebdavTckTestSuite,
)
from webdav_tck.xml_utils import (
    build_propfind_body,
    build_proppatch_body,
    is_wellformed_xml,
    parse_multistatus,
)

if TYPE_CHECKING:
    from webdav_tck.session import WebdavTckSession

NS = "http://example.com/neon/litmus/"


async def test_propfind_d0(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPFIND: Depth 0 PROPFIND on root collection."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Request resourcetype property
    props = [("DAV:", "resourcetype")]
    body = build_propfind_body(props)

    response = await session.client.propfind(
        session.base_path, depth=0, body=body, test_name="propfind_d0"
    )

    if response.status_code not in (200, 207):
        context.context(f"PROPFIND returned {response.status_code}, expected 207")
        return WebdavTckTestResult.FAIL

    # Parse the multistatus response
    try:
        responses = parse_multistatus(response.content)
    except Exception as e:
        context.context(f"Failed to parse PROPFIND response: {e}")
        return WebdavTckTestResult.FAIL

    if not responses:
        context.context("No responses returned from PROPFIND")
        return WebdavTckTestResult.FAIL

    # Check for resourcetype property indicating collection
    for resp in responses:
        if resp.properties and ("DAV:", "resourcetype") in resp.properties:
            return WebdavTckTestResult.OK

    context.warning("resourcetype property not found in response")
    return WebdavTckTestResult.OK


async def test_propfind_invalid(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPFIND: Invalid XML body should return 400."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Send malformed XML
    invalid_body = b"<foo>"

    response = await session.client.propfind(
        session.base_path, depth=0, body=invalid_body, test_name="propfind_invalid"
    )

    if response.status_code != 400:
        context.context(
            f"PROPFIND with malformed XML should return 400, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_propset(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPPATCH: Set custom properties."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Create a resource for property testing
    prop_path = session.make_path("prop")
    content = b"Property test resource\n"

    response = await session.client.put(prop_path, content, test_name="propset_setup")
    if response.status_code != 201:
        context.context(f"PUT failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Set multiple properties
    set_props = {}
    for n in range(10):
        set_props[NS, f"prop{n}"] = f"value{n}"

    body = build_proppatch_body(set_props=set_props)

    response = await session.client.proppatch(prop_path, body, test_name="propset")

    if response.status_code not in (200, 207):
        context.context(f"PROPPATCH failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_propget(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPFIND: Get custom properties."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop")

    # Request the properties we set
    props = [(NS, f"prop{n}") for n in range(10)]
    body = build_propfind_body(props)

    response = await session.client.propfind(
        prop_path, depth=0, body=body, test_name="propget"
    )

    if response.status_code not in (200, 207):
        context.context(f"PROPFIND failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Parse and verify properties
    try:
        responses = parse_multistatus(response.content)
    except Exception as e:
        context.context(f"Failed to parse response: {e}")
        return WebdavTckTestResult.FAIL

    if not responses:
        context.context("No responses returned")
        return WebdavTckTestResult.FAIL

    # Verify property values
    for resp in responses:
        if not resp.properties:
            continue

        for n in range(10):
            prop_key = (NS, f"prop{n}")
            if prop_key in resp.properties:
                value = resp.properties[prop_key]
                if hasattr(value, "text"):
                    value = value.text
                expected = f"value{n}"
                if str(value) != expected:
                    context.warning(
                        f"Property prop{n} has value '{value}', expected '{expected}'"
                    )

    return WebdavTckTestResult.OK


async def test_propmove(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MOVE: Properties persist across MOVE."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop")
    dest_path = session.make_path("prop2")

    # Clean destination
    await session.client.delete(dest_path, test_name="propmove_cleanup")

    # Move the resource
    response = await session.client.move(
        prop_path, dest_path, overwrite=False, test_name="propmove"
    )

    if response.status_code not in (200, 201, 204):
        context.context(f"MOVE failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_propdeletes(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPPATCH: Remove properties."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop2")

    # Remove first 5 properties
    remove_props = [(NS, f"prop{n}") for n in range(5)]
    body = build_proppatch_body(remove_props=remove_props)

    response = await session.client.proppatch(prop_path, body, test_name="propdeletes")

    if response.status_code not in (200, 207):
        context.context(f"PROPPATCH remove failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_propreplace(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPPATCH: Replace property values."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop2")

    # Replace remaining properties (5-9) with new values
    set_props = {}
    for n in range(5, 10):
        set_props[NS, f"prop{n}"] = f"newvalue{n}"

    body = build_proppatch_body(set_props=set_props)

    response = await session.client.proppatch(prop_path, body, test_name="propreplace")

    if response.status_code not in (200, 207):
        context.context(f"PROPPATCH replace failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_propnullns(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPPATCH: Property with null namespace."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop2")

    # Set property with null namespace
    set_props = {("", "nonamespace"): "randomvalue"}
    body = build_proppatch_body(set_props=set_props)

    response = await session.client.proppatch(prop_path, body, test_name="propnullns")

    if response.status_code not in (200, 207):
        context.warning(
            f"PROPPATCH with null namespace returned {response.status_code}"
        )
        # This is a warning, not a failure - some servers may not support it
        return WebdavTckTestResult.OK

    return WebdavTckTestResult.OK


async def test_prophighunicode(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPPATCH: Property with high Unicode character (U+10000)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop2")

    # High Unicode character (U+10000 = UTF-8: F0 90 80 80)
    high_unicode = "\U00010000"  # Unicode character U+10000
    set_props = {(NS, "high-unicode"): high_unicode}
    body = build_proppatch_body(set_props=set_props)

    response = await session.client.proppatch(
        prop_path, body, test_name="prophighunicode"
    )

    if response.status_code not in (200, 207):
        context.warning(f"PROPPATCH with high Unicode returned {response.status_code}")
        # This is a warning - not all servers support high Unicode
        return WebdavTckTestResult.OK

    return WebdavTckTestResult.OK


async def test_propwformed(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPFIND: Response is well-formed XML."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop2")

    # Request all properties
    body = build_propfind_body(allprop=True)

    response = await session.client.propfind(
        prop_path, depth=0, body=body, test_name="propwformed"
    )

    if response.status_code not in (200, 207):
        context.context(f"PROPFIND failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Check if response is well-formed XML
    if not is_wellformed_xml(response.content):
        context.context("PROPFIND response is not well-formed XML")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_propget_lastmodified(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPFIND: Get getlastmodified live property."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    prop_path = session.make_path("prop2")

    # Request getlastmodified property
    props = [("DAV:", "getlastmodified")]
    body = build_propfind_body(props)

    response = await session.client.propfind(
        prop_path, depth=0, body=body, test_name="propget_lastmodified"
    )

    if response.status_code not in (200, 207):
        context.context(f"PROPFIND failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Parse response
    try:
        responses = parse_multistatus(response.content)
    except Exception as e:
        context.context(f"Failed to parse response: {e}")
        return WebdavTckTestResult.FAIL

    # Verify getlastmodified is present
    for resp in responses:
        if resp.properties and ("DAV:", "getlastmodified") in resp.properties:
            value = resp.properties["DAV:", "getlastmodified"]
            if hasattr(value, "text"):
                value = value.text
            if value:
                return WebdavTckTestResult.OK
            context.warning("getlastmodified property is empty")
            return WebdavTckTestResult.OK

    context.warning("getlastmodified property not found")
    return WebdavTckTestResult.OK


def create_props_suite(session: WebdavTckSession) -> WebdavTckTestSuite:
    """Create the properties test suite.

    Args:
        session: WebdavTck session

    Returns:
        Test suite with property tests
    """
    suite = WebdavTckTestSuite("props", session=session)

    # Add tests in order
    tests = [
        test_propfind_d0,
        test_propfind_invalid,
        test_propset,
        test_propget,
        test_propmove,
        test_propdeletes,
        test_propreplace,
        test_propnullns,
        test_prophighunicode,
        test_propwformed,
        test_propget_lastmodified,
    ]

    for test_func in tests:
        suite.add_test(test_func)

    return suite



"""Locking test suite.

Ports src/locks.c from the original litmus C implementation.
Tests WebDAV Class 2 locking functionality including exclusive/shared locks,
lock discovery, refresh, and conditional requests.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from webdav_tck.framework import (
    WebdavTckContext,
    WebdavTckTestResult,
    WebdavTckTestSuite,
)
from webdav_tck.xml_utils import (
    build_if_header,
    build_lock_body,
    build_propfind_body,
    build_proppatch_body,
    parse_lock_response,
)

if TYPE_CHECKING:
    from webdav_tck.session import WebdavTckSession

NS = "http://webdav.org/neon/litmus/"


async def test_options(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """OPTIONS: Check server capabilities."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    response = await session.client.options(session.base_path, test_name="options")

    if response.status_code not in (200, 204):
        context.context(f"OPTIONS returned {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Already checked in session.discover_capabilities()
    return WebdavTckTestResult.OK


async def test_precond(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Precondition: Check server supports Class 2 (locking)."""
    if not session.class2:
        context.context(
            "Locking tests skipped, server does not claim Class 2 compliance"
        )
        return WebdavTckTestResult.SKIPREST

    return WebdavTckTestResult.OK


async def test_init_locks(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Initialize lock resources."""
    # Just a marker test - actual initialization happens in session
    return WebdavTckTestResult.OK


async def test_put(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PUT: Create resources for locking tests."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Create main resource
    lockme = session.make_path("lockme")
    content = b"This is a test file for locking.\n"

    response = await session.client.put(lockme, content, test_name="put_lockme")
    if response.status_code != 201:
        context.context(f"PUT lockme returned {response.status_code}, expected 201")
        return WebdavTckTestResult.FAILHARD

    # Create second resource (not locked)
    notlocked = session.make_path("notlocked")
    response = await session.client.put(notlocked, content, test_name="put_notlocked")
    if response.status_code != 201:
        context.context(f"PUT notlocked returned {response.status_code}, expected 201")
        return WebdavTckTestResult.FAILHARD

    return WebdavTckTestResult.OK


async def test_lock_excl(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Take out exclusive lock on resource."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Build lock request body
    body = build_lock_body(
        scope="exclusive",
        lock_type="write",
        owner="litmus test suite",
    )

    # Send LOCK request with Depth: 0
    response = await session.client.lock(
        lockme, body, depth="0", timeout="Second-3600", test_name="lock_excl"
    )

    if response.status_code not in (200, 201):
        context.context(f"LOCK failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Parse lock token from response
    try:
        lock_token, lock_scope = parse_lock_response(response.content)
    except Exception as e:
        context.context(f"Failed to parse LOCK response: {e}")
        return WebdavTckTestResult.FAIL

    if not lock_token:
        context.context("No lock token returned from LOCK")
        return WebdavTckTestResult.FAIL

    if lock_scope != "exclusive":
        context.context(f"Requested exclusive lock but got {lock_scope}")
        return WebdavTckTestResult.FAIL

    # Store lock token for later tests
    session.locks["lockme"] = lock_token

    return WebdavTckTestResult.OK


async def test_discover(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PROPFIND: Discover lock information."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Use PROPFIND with lockdiscovery property
    props = [("DAV:", "lockdiscovery")]
    body = build_propfind_body(props)

    response = await session.client.propfind(
        lockme, depth=0, body=body, test_name="discover"
    )

    if response.status_code not in (200, 207):
        context.context(f"PROPFIND lockdiscovery failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # TODO: Parse and verify lockdiscovery property contains our lock token
    # For now, just verify the request succeeded
    return WebdavTckTestResult.OK


async def test_refresh(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Refresh existing lock."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Refresh lock by sending LOCK with If: header
    response = await session.client.lock(
        lockme,
        None,
        depth="0",
        timeout="Second-3600",
        if_header=f"(<{lock_token}>)",
        test_name="refresh",
    )

    if response.status_code != 200:
        context.context(f"LOCK refresh failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_notowner_modify(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Non-owner: Verify non-owner cannot modify locked resource."""
    if not session.client2:
        context.context("Second client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")
    notlocked = session.make_path("notlocked")

    # Try DELETE with second session (no lock token) - should fail
    response = await session.client2.delete(lockme, test_name="notowner_delete")
    if response.status_code not in (400, 423):
        context.context(
            f"DELETE of locked resource should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"DELETE failed with {response.status_code} not 423")

    # Try MOVE - should fail
    whocares = session.make_path("whocares")
    response = await session.client2.move(
        lockme, whocares, overwrite=False, test_name="notowner_move"
    )
    if response.status_code not in (400, 423):
        context.context(
            f"MOVE of locked resource should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"MOVE failed with {response.status_code} not 423")

    # Try COPY onto locked resource - should fail
    response = await session.client2.copy(
        notlocked, lockme, overwrite=True, depth="0", test_name="notowner_copy"
    )
    if response.status_code not in (400, 423):
        context.context(
            f"COPY onto locked resource should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"COPY failed with {response.status_code} not 423")

    # Try PROPPATCH - should fail
    set_props = {(NS, "random"): "foobar"}
    body = build_proppatch_body(set_props=set_props)

    response = await session.client2.proppatch(
        lockme, body, test_name="notowner_proppatch"
    )
    if response.status_code not in (400, 423):
        context.context(
            f"PROPPATCH of locked resource should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"PROPPATCH failed with {response.status_code} not 423")

    # Try PUT - should fail
    content = b"Trying to modify locked resource\n"
    response = await session.client2.put(lockme, content, test_name="notowner_put")
    if response.status_code not in (400, 423):
        context.context(
            f"PUT on locked resource should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"PUT failed with {response.status_code} not 423")

    return WebdavTckTestResult.OK


async def test_notowner_lock(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Non-owner: Verify non-owner cannot unlock or lock resource."""
    if not session.client2:
        context.context("Second client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Try UNLOCK with bogus token - should fail
    bogus_token = "opaquelocktoken:foobar"
    response = await session.client2.unlock(
        lockme, bogus_token, test_name="notowner_unlock"
    )
    if response.status_code < 400:
        context.context("UNLOCK with bogus lock token should fail")
        return WebdavTckTestResult.FAIL

    # Try to take out another exclusive lock - should fail
    body = build_lock_body(
        scope="exclusive",
        lock_type="write",
        owner="notowner lock",
    )

    response = await session.client2.lock(
        lockme, body, depth="0", timeout="Second-3600", test_name="notowner_lock"
    )
    if response.status_code not in (400, 423):
        context.context(
            f"LOCK on locked resource should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"LOCK failed with {response.status_code} not 423")

    return WebdavTckTestResult.OK


async def test_owner_modify(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Owner: Verify lock owner can modify resource."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # PUT with lock token should succeed
    if_header = build_if_header(lock_token)
    content = b"Modified by lock owner\n"

    response = await session.client.put(
        lockme, content, if_header=if_header, test_name="owner_put"
    )
    if response.status_code not in (200, 201, 204):
        context.context(
            f"PUT on locked resource by owner failed: {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # PROPPATCH with lock token should succeed
    set_props = {(NS, "random"): "foobar"}
    body = build_proppatch_body(set_props=set_props)

    response = await session.client.proppatch(
        lockme, body, if_header=if_header, test_name="owner_proppatch"
    )
    if response.status_code not in (200, 207):
        context.context(
            f"PROPPATCH on locked resource by owner failed: {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_copy(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """COPY: Verify locks don't follow COPY operations."""
    if not session.client2:
        context.context("Second client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")
    copydest = session.make_path("lockme-copydest")

    # Clean destination
    await session.client2.delete(copydest, test_name="copy_cleanup")

    # Copy locked resource (should succeed, and copy should not be locked)
    response = await session.client2.copy(
        lockme, copydest, overwrite=True, depth="0", test_name="copy"
    )
    if response.status_code not in (200, 201, 204):
        context.context(f"COPY of locked resource failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Verify copy is not locked by trying to delete it without lock token
    response = await session.client2.delete(copydest, test_name="copy_verify")
    if response.status_code not in (200, 204):
        context.context(
            f"Copied resource appears to be locked: DELETE returned {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_cond_put(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Conditional PUT: PUT with valid lock token and etag."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Get current ETag
    etag = session.etags.get("lockme")
    if not etag:
        # Try to get it
        response = await session.client.get(lockme, test_name="cond_put_get")
        if response.status_code == 200:
            etag = response.headers.get("etag")
            if etag:
                session.etags["lockme"] = etag

    if not etag:
        context.warning("No ETag available, skipping conditional PUT test")
        return WebdavTckTestResult.OK

    # Build If: header with lock token and etag
    if_header = f"(<{lock_token}> [{etag}])"
    content = b"Conditional PUT with lock and etag\n"

    response = await session.client.put(
        lockme, content, if_header=if_header, test_name="cond_put"
    )
    if response.status_code not in (200, 204):
        context.context(
            f"Conditional PUT with lock and etag failed: {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # Update ETag if returned
    new_etag = response.headers.get("etag")
    if new_etag:
        session.etags["lockme"] = new_etag

    return WebdavTckTestResult.OK


async def test_fail_cond_put(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Conditional PUT: PUT with bogus lock token should fail."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")
    etag = session.etags.get("lockme")

    if not etag:
        context.warning("No ETag available, skipping test")
        return WebdavTckTestResult.OK

    # Build If: header with bogus lock token and valid etag
    if_header = f"(<DAV:no-lock> [{etag}])"
    content = b"This should fail\n"

    response = await session.client.put(
        lockme, content, if_header=if_header, test_name="fail_cond_put"
    )

    if response.status_code < 400:
        context.context("Conditional PUT with invalid lock token should fail")
        return WebdavTckTestResult.FAIL

    if response.status_code == 400:
        context.context(
            "Conditional PUT with invalid lock token got 400 (should be 412)"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 412:
        context.warning(f"PUT failed with {response.status_code} not 412")

    return WebdavTckTestResult.OK


async def test_cond_put_with_not(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Conditional PUT: PUT with lock token and Not clause."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Build If: header: (lock-token) (Not bogus-token)
    if_header = f"(<{lock_token}>) (Not <DAV:no-lock>)"
    content = b"Conditional PUT with Not clause\n"

    response = await session.client.put(
        lockme, content, if_header=if_header, test_name="cond_put_with_not"
    )
    if response.status_code not in (200, 204):
        context.context(
            f"PUT with conditional (Not <DAV:no-lock>) failed: {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # Update ETag
    etag = response.headers.get("etag")
    if etag:
        session.etags["lockme"] = etag

    return WebdavTckTestResult.OK


async def test_cond_put_corrupt_token(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Conditional PUT: PUT with corrupted lock token should fail."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Corrupt the lock token
    corrupt_token = lock_token + "x"
    if_header = f"(<{corrupt_token}>) (Not <DAV:no-lock>)"
    content = b"This should fail\n"

    response = await session.client.put(
        lockme, content, if_header=if_header, test_name="cond_put_corrupt"
    )
    if response.status_code < 400:
        context.context("Conditional PUT with invalid lock token should fail")
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"PUT failed with {response.status_code} not 423")

    return WebdavTckTestResult.OK


async def test_unlock(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """UNLOCK: Release the lock."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    response = await session.client.unlock(lockme, lock_token, test_name="unlock")
    if response.status_code != 204:
        context.context(f"UNLOCK failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Remove lock token
    del session.locks["lockme"]

    return WebdavTckTestResult.OK


async def test_fail_cond_put_unlocked(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Conditional PUT: PUT with bogus token on unlocked resource."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Try conditional PUT with bogus token on unlocked resource
    if_header = "(<DAV:no-lock>)"
    content = b"This should fail\n"

    response = await session.client.put(
        lockme, content, if_header=if_header, test_name="fail_cond_put_unlocked"
    )
    if response.status_code < 400:
        context.context("Conditional PUT with invalid lock token should fail")
        return WebdavTckTestResult.FAIL

    if response.status_code == 400:
        context.context(
            "Conditional PUT with invalid lock token got 400 (should be 412)"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 412:
        context.warning(f"PUT failed with {response.status_code} not 412")

    return WebdavTckTestResult.OK


async def test_lock_shared(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Take out shared lock on resource."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Build lock request body for shared lock
    body = build_lock_body(
        scope="shared",
        lock_type="write",
        owner="litmus test suite",
    )

    response = await session.client.lock(
        lockme, body, depth="0", timeout="Second-3600", test_name="lock_shared"
    )

    if response.status_code not in (200, 201):
        context.context(f"LOCK failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Parse lock token
    try:
        lock_token, lock_scope = parse_lock_response(response.content)
    except Exception as e:
        context.context(f"Failed to parse LOCK response: {e}")
        return WebdavTckTestResult.FAIL

    if not lock_token:
        context.context("No lock token returned from LOCK")
        return WebdavTckTestResult.FAIL

    if lock_scope != "shared":
        context.context(f"Requested shared lock but got {lock_scope}")
        return WebdavTckTestResult.FAIL

    session.locks["lockme"] = lock_token

    return WebdavTckTestResult.OK


async def test_double_sharedlock(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Take out second shared lock on resource."""
    if not session.client2:
        context.context("Second client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockme")
    if not lock_token:
        context.context("No lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme = session.make_path("lockme")

    # Take out second shared lock with client2
    body = build_lock_body(
        scope="shared",
        lock_type="write",
        owner="litmus: notowner_sharedlock",
    )

    response = await session.client2.lock(
        lockme, body, depth="0", timeout="Second-3600", test_name="double_sharedlock"
    )

    if response.status_code not in (200, 201):
        context.context(f"Second shared LOCK failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Parse second lock token
    try:
        lock_token2, _lock_scope = parse_lock_response(response.content)
    except Exception as e:
        context.context(f"Failed to parse LOCK response: {e}")
        return WebdavTckTestResult.FAIL

    if not lock_token2:
        context.context("No lock token returned from second LOCK")
        return WebdavTckTestResult.FAIL

    # Unlock the second lock
    response = await session.client2.unlock(
        lockme, lock_token2, test_name="double_sharedlock_unlock"
    )
    if response.status_code != 204:
        context.context(f"UNLOCK of second shared lock failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_prep_collection(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Prepare collection for collection locking tests."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Clear existing lock if any
    if "lockme" in session.locks:
        del session.locks["lockme"]

    # Create collection
    lockcoll = session.make_path("lockcoll")
    response = await session.client.mkcol(lockcoll, test_name="prep_collection")

    if response.status_code != 201:
        context.context(f"MKCOL failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    return WebdavTckTestResult.OK


async def test_lock_collection(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Take out depth infinity lock on collection."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lockcoll = session.make_path("lockcoll")

    # Lock collection with Depth: infinity
    body = build_lock_body(
        scope="exclusive",
        lock_type="write",
        owner="litmus test suite",
    )

    response = await session.client.lock(
        lockcoll,
        body,
        depth="infinity",
        timeout="Second-3600",
        test_name="lock_collection",
    )

    if response.status_code not in (200, 201):
        context.context(f"LOCK collection failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Parse lock token
    try:
        lock_token, _lock_scope = parse_lock_response(response.content)
    except Exception as e:
        context.context(f"Failed to parse LOCK response: {e}")
        return WebdavTckTestResult.FAIL

    if not lock_token:
        context.context("No lock token returned")
        return WebdavTckTestResult.FAIL

    session.locks["lockcoll"] = lock_token

    # Create a resource in the locked collection
    lockme_txt = session.make_path("lockcoll", "lockme.txt")
    content = b"Resource in locked collection\n"
    if_header = build_if_header(lock_token)

    response = await session.client.put(
        lockme_txt, content, if_header=if_header, test_name="lock_collection_put"
    )
    if response.status_code not in (201, 204):
        context.context(f"PUT in locked collection failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Store the member path for later tests
    session.locks["lockme"] = lock_token  # Use same token for member

    return WebdavTckTestResult.OK


async def test_owner_modify_coll(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Owner: Verify lock owner can modify collection member."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockcoll")
    if not lock_token:
        context.context("No collection lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme_txt = session.make_path("lockcoll", "lockme.txt")

    # PUT with lock token should succeed
    if_header = build_if_header(lock_token)
    content = b"Modified by collection lock owner\n"

    response = await session.client.put(
        lockme_txt, content, if_header=if_header, test_name="owner_put_coll"
    )
    if response.status_code not in (200, 201, 204):
        context.context(
            f"PUT on locked collection member by owner failed: {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # PROPPATCH with lock token should succeed
    set_props = {(NS, "random"): "foobar"}
    body = build_proppatch_body(set_props=set_props)

    response = await session.client.proppatch(
        lockme_txt, body, if_header=if_header, test_name="owner_proppatch_coll"
    )
    if response.status_code not in (200, 207):
        context.context(
            f"PROPPATCH on locked collection member by owner failed: {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_notowner_modify_coll(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """Non-owner: Verify non-owner cannot modify collection member."""
    if not session.client2:
        context.context("Second client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockcoll")
    if not lock_token:
        context.context("No collection lock token available")
        return WebdavTckTestResult.FAILHARD

    lockme_txt = session.make_path("lockcoll", "lockme.txt")
    notlocked = session.make_path("notlocked")

    # Try DELETE with second session (no lock token) - should fail
    response = await session.client2.delete(
        lockme_txt, test_name="notowner_delete_coll"
    )
    if response.status_code not in (400, 423):
        context.context(
            f"DELETE of locked collection member should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"DELETE failed with {response.status_code} not 423")

    # Try PUT - should fail
    content = b"Trying to modify locked collection member\n"
    response = await session.client2.put(
        lockme_txt, content, test_name="notowner_put_coll"
    )
    if response.status_code not in (400, 423):
        context.context(
            f"PUT on locked collection member should fail, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 423:
        context.warning(f"PUT failed with {response.status_code} not 423")

    return WebdavTckTestResult.OK


async def test_indirect_refresh(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Indirectly refresh collection lock via member."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockcoll")
    if not lock_token:
        context.context("No collection lock token available")
        return WebdavTckTestResult.FAILHARD

    # Refresh the collection lock using the member path
    lockme_txt = session.make_path("lockcoll", "lockme.txt")

    response = await session.client.lock(
        lockme_txt,
        None,
        depth="0",
        timeout="Second-3600",
        if_header=f"(<{lock_token}>)",
        test_name="indirect_refresh",
    )

    if response.status_code != 200:
        context.context(f"Indirect refresh LOCK failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_unlock_collection(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """UNLOCK: Release collection lock."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("lockcoll")
    if not lock_token:
        context.context("No collection lock token available")
        return WebdavTckTestResult.FAILHARD

    lockcoll = session.make_path("lockcoll")

    response = await session.client.unlock(
        lockcoll, lock_token, test_name="unlock_collection"
    )
    if response.status_code != 204:
        context.context(f"UNLOCK collection failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Clear locks
    if "lockcoll" in session.locks:
        del session.locks["lockcoll"]
    if "lockme" in session.locks:
        del session.locks["lockme"]

    return WebdavTckTestResult.OK


async def test_unmapped_lock(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """LOCK: Lock on unmapped URL should return 201."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Clear any existing locks
    if "lockme" in session.locks:
        del session.locks["lockme"]

    unmapped = session.make_path("unmapped_url")

    # Lock unmapped URL (creates lock-null resource)
    body = build_lock_body(
        scope="exclusive",
        lock_type="write",
        owner="litmus test suite",
    )

    response = await session.client.lock(
        unmapped, body, depth="0", timeout="Second-3600", test_name="unmapped_lock"
    )

    if response.status_code not in (200, 201):
        context.context(f"LOCK on unmapped URL failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    if response.status_code != 201:
        context.warning(
            f"LOCK on unmapped URL returned {response.status_code} not 201 (RFC4918:S7.3)"
        )

    # Parse and store lock token
    try:
        lock_token, _ = parse_lock_response(response.content)
        if lock_token:
            session.locks["unmapped_url"] = lock_token
    except Exception:
        pass

    return WebdavTckTestResult.OK


async def test_unlock_unmapped(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """UNLOCK: Release lock on unmapped resource."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    lock_token = session.locks.get("unmapped_url")
    if not lock_token:
        # Not a hard failure - maybe previous test was skipped
        return WebdavTckTestResult.OK

    unmapped = session.make_path("unmapped_url")

    response = await session.client.unlock(
        unmapped, lock_token, test_name="unlock_unmapped"
    )
    if response.status_code != 204:
        context.context(f"UNLOCK unmapped resource failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    del session.locks["unmapped_url"]

    return WebdavTckTestResult.OK


def create_locks_suite(session: WebdavTckSession) -> WebdavTckTestSuite:
    """Create the locks test suite.

    Args:
        session: WebdavTck session

    Returns:
        Test suite with locking tests
    """
    suite = WebdavTckTestSuite("locks", session=session)

    # Add tests in order matching locks.c
    tests = [
        test_options,
        test_precond,
        test_init_locks,
        # Exclusive lock tests
        test_put,
        test_lock_excl,
        test_discover,
        test_refresh,
        test_notowner_modify,
        test_notowner_lock,
        test_owner_modify,
        # Check lock persists after modification
        test_notowner_modify,
        test_notowner_lock,
        # Locks don't follow COPY
        test_copy,
        # Conditional PUTs
        test_cond_put,
        test_fail_cond_put,
        test_cond_put_with_not,
        test_cond_put_corrupt_token,
        test_unlock,
        test_fail_cond_put_unlocked,
        # Shared lock tests
        test_lock_shared,
        test_notowner_modify,
        test_notowner_lock,
        test_owner_modify,
        test_double_sharedlock,
        # Check main lock still intact
        test_notowner_modify,
        test_notowner_lock,
        test_unlock,
        # Collection locking
        test_prep_collection,
        test_lock_collection,
        test_owner_modify_coll,
        test_notowner_modify_coll,
        test_refresh,
        test_indirect_refresh,
        test_unlock_collection,
        # Lock-null resources
        test_unmapped_lock,
        test_unlock_unmapped,
    ]

    for test_func in tests:
        suite.add_test(test_func)

    return suite



"""Session management for litmus tests."""

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import ParseResult, urlparse

from typing_extensions import Self

from webdav_tck.client import WebDAVClient
from webdav_tck.xml_utils import extract_dav_header

if TYPE_CHECKING:
    from pathlib import Path


class WebdavTckSession:
    """Manages test session state.

    Corresponds to the global session state in the C implementation
    (i_session, i_session2, i_origin, etc.).
    """

    def __init__(
        self,
        url: str,
        username: str | None = None,
        password: str | None = None,
        verify_ssl: bool = True,
        proxy: str | None = None,
        client_cert: str | None = None,
        debug_log: Path | None = None,
    ):
        """Initialize litmus session.

        Args:
            url: Base URL for WebDAV server
            username: Optional username for authentication
            password: Optional password for authentication
            verify_ssl: Whether to verify SSL certificates
            proxy: Optional proxy URL
            client_cert: Optional path to client certificate
            debug_log: Optional path to debug log file
        """
        self.origin: ParseResult = urlparse(url)
        self.username = username
        self.password = password
        self.verify_ssl = verify_ssl
        self.proxy = proxy
        self.client_cert = client_cert
        self.debug_log = debug_log

        # WebDAV capabilities
        self.class2 = False  # WebDAV Class 2 (locking) support
        self.capabilities: dict[str, bool] = {}

        # Test collection path
        self.base_path = self.origin.path.rstrip("/") + "/litmus"

        # Clients (primary and secondary for multi-user tests)
        self.client: WebDAVClient | None = None
        self.client2: WebDAVClient | None = None

        # State tracking
        self.locks: dict[str, str] = {}  # path -> lock token
        self.etags: dict[str, str] = {}  # path -> etag

    @property
    def hostname(self) -> str:
        """Get hostname from origin."""
        return self.origin.hostname or "localhost"

    @property
    def port(self) -> int:
        """Get port from origin."""
        return self.origin.port or (443 if self.origin.scheme == "https" else 80)

    @property
    def base_url(self) -> str:
        """Get base URL for the server."""
        return f"{self.origin.scheme}://{self.origin.netloc}"

    async def __aenter__(self) -> Self:
        """Async context manager entry - create clients."""
        await self.begin()
        return self

    async def __aexit__(self, *args: object) -> None:
        """Async context manager exit - close clients."""
        await self.finish()

    async def begin(self) -> None:
        """Initialize session and create WebDAV clients.

        Creates primary and secondary clients for testing.
        """
        # Create primary client
        self.client = WebDAVClient(
            base_url=self.base_url,
            username=self.username,
            password=self.password,
            verify_ssl=self.verify_ssl,
            proxy=self.proxy,
            client_cert=self.client_cert,
            litmus_header="X-WebdavTck",
        )
        # Manually enter context for lifecycle management
        self.client = await self.client.__aenter__()  # noqa: PLC2801

        # Create secondary client for multi-user tests (locks)
        self.client2 = WebDAVClient(
            base_url=self.base_url,
            username=self.username,
            password=self.password,
            verify_ssl=self.verify_ssl,
            proxy=self.proxy,
            client_cert=self.client_cert,
            litmus_header="X-WebdavTck-Second",
        )
        # Manually enter context for lifecycle management
        self.client2 = await self.client2.__aenter__()  # noqa: PLC2801

    async def finish(self) -> None:
        """Clean up session and close clients."""
        if self.client:
            await self.client.close()
        if self.client2:
            await self.client2.close()

    async def discover_capabilities(self) -> None:
        """Discover WebDAV server capabilities via OPTIONS.

        Sets self.class2 and self.capabilities based on DAV header.
        """
        if not self.client:
            msg = "Session not initialized - call begin() first"
            raise RuntimeError(msg)

        response = await self.client.options(self.base_path, test_name="options")

        # Parse DAV header
        dav_header = response.headers.get("DAV", "")
        if dav_header:
            self.capabilities = extract_dav_header(dav_header)
            # WebDAV Class 2 means locking support
            self.class2 = "2" in self.capabilities

    async def create_test_collection(self) -> None:
        """Create the litmus test collection.

        Creates the base collection where all test resources will be created.
        """
        if not self.client:
            msg = "Session not initialized - call begin() first"
            raise RuntimeError(msg)

        # Delete if exists (cleanup from previous run)
        try:
            await self.client.delete(self.base_path, test_name="cleanup")
        except Exception:
            pass  # Ignore if doesn't exist

        # Create fresh collection
        response = await self.client.mkcol(self.base_path, test_name="setup")
        if response.status_code not in (201, 405):  # 405 = already exists
            msg = f"Failed to create test collection: {response.status_code}"
            raise RuntimeError(msg)

    async def cleanup_test_collection(self) -> None:
        """Remove the litmus test collection.

        Cleans up all test resources created during the test run.
        """
        if not self.client:
            return

        try:
            await self.client.delete(self.base_path, test_name="cleanup")
        except Exception:
            pass  # Ignore errors during cleanup

    def store_lock(self, path: str, token: str) -> None:
        """Store a lock token for a resource.

        Args:
            path: Resource path
            token: Lock token
        """
        self.locks[path] = token

    def get_lock(self, path: str) -> str | None:
        """Get stored lock token for a resource.

        Args:
            path: Resource path

        Returns:
            Lock token or None if not locked
        """
        return self.locks.get(path)

    def remove_lock(self, path: str) -> None:
        """Remove stored lock token.

        Args:
            path: Resource path
        """
        self.locks.pop(path, None)

    def store_etag(self, path: str, etag: str) -> None:
        """Store ETag for a resource.

        Args:
            path: Resource path
            etag: ETag value
        """
        self.etags[path] = etag

    def get_etag(self, path: str) -> str | None:
        """Get stored ETag for a resource.

        Args:
            path: Resource path

        Returns:
            ETag or None if not stored
        """
        return self.etags.get(path)

    def make_path(self, *parts: str) -> str:
        """Make a path within the test collection.

        Args:
            *parts: Path components

        Returns:
            Full path relative to base
        """
        return "/".join([self.base_path] + list(parts))



"""WebDAV test suites."""

from __future__ import annotations

from webdav_tck.suites.basic import create_basic_suite
from webdav_tck.suites.copymove import create_copymove_suite
from webdav_tck.suites.locks import create_locks_suite
from webdav_tck.suites.props import create_props_suite

__all__ = [
    "create_basic_suite",
    "create_copymove_suite",
    "create_locks_suite",
    "create_props_suite",
]



"""XML utilities for WebDAV protocol."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from lxml import etree

if TYPE_CHECKING:
    from lxml.etree import _Element

# WebDAV namespaces
NS_DAV = "DAV:"
NS_MAP = {"D": NS_DAV}


@dataclass
class PropValue:
    """A property value with namespace."""

    namespace: str
    name: str
    value: Any


@dataclass
class MultiStatusResponse:
    """Response element from multistatus."""

    href: str
    status: int | None = None
    properties: dict[tuple[str, str], Any] | None = None  # (namespace, name) -> value
    error: str | None = None


@dataclass
class LockInfo:
    """Information about a lock."""

    token: str
    timeout: str | None = None
    owner: str | None = None
    scope: str = "exclusive"  # exclusive or shared
    depth: str = "0"


def build_propfind_body(
    props: list[tuple[str, str]] | None = None,
    allprop: bool = False,
    propname: bool = False,
) -> bytes:
    """Build PROPFIND XML request body.

    Args:
        props: List of (namespace, name) tuples for specific properties
        allprop: Request all properties
        propname: Request only property names

    Returns:
        XML as bytes
    """
    root = etree.Element("{DAV:}propfind", nsmap=NS_MAP)

    if allprop:
        etree.SubElement(root, "{DAV:}allprop")
    elif propname:
        etree.SubElement(root, "{DAV:}propname")
    elif props:
        prop = etree.SubElement(root, "{DAV:}prop")
        for namespace, name in props:
            etree.SubElement(prop, f"{{{namespace}}}{name}")
    else:
        # Default: allprop
        etree.SubElement(root, "{DAV:}allprop")

    return etree.tostring(root, xml_declaration=True, encoding="UTF-8")


def build_proppatch_body(
    set_props: dict[tuple[str, str], str | _Element] | None = None,
    remove_props: list[tuple[str, str]] | None = None,
) -> bytes:
    """Build PROPPATCH XML request body.

    Args:
        set_props: Properties to set as {(namespace, name): value}
        remove_props: Properties to remove as [(namespace, name)]

    Returns:
        XML as bytes
    """
    root = etree.Element("{DAV:}propertyupdate", nsmap=NS_MAP)

    # Add set operations
    if set_props:
        set_elem = etree.SubElement(root, "{DAV:}set")
        prop = etree.SubElement(set_elem, "{DAV:}prop")

        for (namespace, name), value in set_props.items():
            prop_elem = etree.SubElement(prop, f"{{{namespace}}}{name}")
            if isinstance(value, etree._Element):
                prop_elem.append(value)
            else:
                prop_elem.text = str(value)

    # Add remove operations
    if remove_props:
        remove_elem = etree.SubElement(root, "{DAV:}remove")
        prop = etree.SubElement(remove_elem, "{DAV:}prop")

        for namespace, name in remove_props:
            etree.SubElement(prop, f"{{{namespace}}}{name}")

    return etree.tostring(root, xml_declaration=True, encoding="UTF-8")


def build_lock_body(
    owner: str, scope: str = "exclusive", lock_type: str = "write"
) -> bytes:
    """Build LOCK XML request body.

    Args:
        owner: Lock owner identifier (URL or text)
        scope: Lock scope (exclusive or shared)
        lock_type: Lock type (write)

    Returns:
        XML as bytes
    """
    root = etree.Element("{DAV:}lockinfo", nsmap=NS_MAP)

    # Lock scope
    lockscope = etree.SubElement(root, "{DAV:}lockscope")
    etree.SubElement(lockscope, f"{{DAV:}}{scope}")

    # Lock type
    locktype_elem = etree.SubElement(root, "{DAV:}locktype")
    etree.SubElement(locktype_elem, f"{{DAV:}}{lock_type}")

    # Owner
    owner_elem = etree.SubElement(root, "{DAV:}owner")
    if owner.startswith("http://") or owner.startswith("https://"):
        href = etree.SubElement(owner_elem, "{DAV:}href")
        href.text = owner
    else:
        owner_elem.text = owner

    return etree.tostring(root, xml_declaration=True, encoding="UTF-8")


def parse_multistatus(xml_data: bytes) -> list[MultiStatusResponse]:
    """Parse WebDAV multistatus XML response.

    Args:
        xml_data: XML response body

    Returns:
        List of response elements
    """
    try:
        root = etree.fromstring(xml_data)
    except etree.XMLSyntaxError as e:
        msg = f"Invalid XML: {e}"
        raise ValueError(msg) from e

    responses: list[MultiStatusResponse] = []

    # Find all response elements
    for response_elem in root.findall(".//{DAV:}response"):
        href_elem = response_elem.find("{DAV:}href")
        if href_elem is None or href_elem.text is None:
            continue

        href = href_elem.text

        # Check for status (for simple responses)
        status_elem = response_elem.find("{DAV:}status")
        status = None
        if status_elem is not None and status_elem.text:
            # Parse "HTTP/1.1 200 OK" -> 200
            parts = status_elem.text.split()
            if len(parts) >= 2:
                try:
                    status = int(parts[1])
                except ValueError:
                    pass

        # Parse propstat elements
        properties: dict[tuple[str, str], Any] = {}
        error = None

        for propstat in response_elem.findall("{DAV:}propstat"):
            propstat_status_elem = propstat.find("{DAV:}status")
            if propstat_status_elem is not None and propstat_status_elem.text:
                parts = propstat_status_elem.text.split()
                if len(parts) >= 2:
                    try:
                        propstat_status = int(parts[1])
                    except ValueError:
                        continue

                    # Only process successful properties
                    if 200 <= propstat_status < 300:
                        prop_elem = propstat.find("{DAV:}prop")
                        if prop_elem is not None:
                            for prop in prop_elem:
                                # Convert tag to string to handle QName objects
                                tag_str = str(prop.tag)
                                namespace = (
                                    tag_str.split("}", maxsplit=1)[0][1:]
                                    if "}" in tag_str
                                    else ""
                                )
                                name = (
                                    tag_str.split("}")[1] if "}" in tag_str else tag_str
                                )
                                value = _extract_prop_value(prop)
                                properties[namespace, name] = value

        responses.append(
            MultiStatusResponse(
                href=href, status=status, properties=properties, error=error
            )
        )

    return responses


def _extract_prop_value(prop_elem: _Element) -> Any:
    """Extract value from property element.

    Args:
        prop_elem: Property element

    Returns:
        Property value (string or element)
    """
    # If property has children, return the element itself
    if len(prop_elem) > 0:
        return prop_elem

    # Otherwise return text content
    return prop_elem.text or ""


def parse_lock_response(xml_data: bytes) -> LockInfo:
    """Parse LOCK response to extract token and timeout.

    Args:
        xml_data: XML response body

    Returns:
        Lock information
    """
    try:
        root = etree.fromstring(xml_data)
    except etree.XMLSyntaxError as e:
        msg = f"Invalid XML: {e}"
        raise ValueError(msg) from e

    # Find lock token
    token_elem = root.find(".//{DAV:}locktoken/{DAV:}href")
    if token_elem is None or token_elem.text is None:
        msg = "No lock token found in response"
        raise ValueError(msg)

    token = token_elem.text

    # Find timeout
    timeout_elem = root.find(".//{DAV:}lockdiscovery//{DAV:}timeout")
    timeout = timeout_elem.text if timeout_elem is not None else None

    # Find owner
    owner_elem = root.find(".//{DAV:}lockdiscovery//{DAV:}owner")
    owner = None
    if owner_elem is not None:
        owner_href = owner_elem.find("{DAV:}href")
        if owner_href is not None and owner_href.text:
            owner = owner_href.text
        else:
            owner = owner_elem.text

    # Find scope
    scope = "exclusive"
    if root.find(".//{DAV:}lockscope/{DAV:}shared") is not None:
        scope = "shared"

    # Find depth
    depth_elem = root.find(".//{DAV:}lockdiscovery//{DAV:}depth")
    depth = depth_elem.text if depth_elem is not None and depth_elem.text else "0"

    return LockInfo(token=token, timeout=timeout, owner=owner, scope=scope, depth=depth)


def extract_dav_header(dav_header: str) -> dict[str, bool]:
    """Parse DAV header to extract capabilities.

    Args:
        dav_header: Value of DAV header

    Returns:
        Dictionary of capabilities (e.g., {"1": True, "2": True})
    """
    capabilities = {}
    for part in dav_header.split(","):
        part = part.strip()
        if part:
            capabilities[part] = True
    return capabilities


def build_if_header(
    lock_token: str | list[tuple[str | None, list[str]]], etag: str | None = None
) -> str:
    """Build If header for conditional requests.

    Args:
        lock_token: Either a simple lock token string, or a list of (resource, [tokens/etags]) tuples
                   for complex conditions
        etag: Optional ETag for simple case

    Returns:
        If header value

    Examples:
        ("token1",) -> "(<token1>)"
        ("token1", "etag1") -> "(<token1> [etag1])"
        [("path1", ["token1"]), ("path2", ["token2"])] -> "<path1> (<token1>) <path2> (<token2>)"
    """
    # Simple case: single lock token with optional etag
    if isinstance(lock_token, str):
        if etag:
            return f"(<{lock_token}> [{etag}])"
        return f"(<{lock_token}>)"

    # Complex case: list of conditions
    parts: list[str] = []
    for resource, tokens in lock_token:
        if resource:
            parts.append(f"<{resource}>")

        token_parts: list[str] = []
        for token in tokens:
            if token.startswith('"'):
                # ETag
                token_parts.append(f"[{token}]")
            else:
                # Lock token
                token_parts.append(f"<{token}>")

        parts.append(f"({' '.join(token_parts)})")

    return " ".join(parts)


def parse_lock_response(xml_data: bytes) -> tuple[str | None, str]:
    """Parse LOCK response to extract lock token and scope.

    Args:
        xml_data: XML response body from LOCK request

    Returns:
        Tuple of (lock_token, scope) where scope is "exclusive" or "shared"
    """
    try:
        root = etree.fromstring(xml_data)
    except etree.XMLSyntaxError:
        return None, "exclusive"

    # Find lock token - typically in activelock/locktoken/href
    token_elem = root.find(".//{DAV:}locktoken/{DAV:}href")
    if token_elem is None:
        # Try alternative path
        token_elem = root.find(".//{DAV:}activelock/{DAV:}locktoken/{DAV:}href")

    token = None
    if token_elem is not None and token_elem.text:
        token = token_elem.text.strip()

    # Find scope
    scope = "exclusive"
    if root.find(".//{DAV:}lockscope/{DAV:}shared") is not None:
        scope = "shared"

    return token, scope


def is_wellformed_xml(xml_data: bytes) -> bool:
    """Check if XML is well-formed.

    Args:
        xml_data: XML data to check

    Returns:
        True if well-formed, False otherwise
    """
    try:
        etree.fromstring(xml_data)
        return True
    except etree.XMLSyntaxError:
        return False



"""COPY and MOVE operations test suite.

Ports src/copymove.c from the original litmus C implementation.
Tests the server's implementation of COPY and MOVE methods under various conditions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from webdav_tck.framework import (
    WebdavTckContext,
    WebdavTckTestResult,
    WebdavTckTestSuite,
)

if TYPE_CHECKING:
    from webdav_tck.session import WebdavTckSession


async def test_copy_simple(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """COPY: Simple resource copy to new destination."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    src = session.make_path("copysrc")
    dest = session.make_path("copydest")

    # Create source resource
    content = b"This is the copy source file.\n"
    response = await session.client.put(src, content, test_name="copy_simple_setup")
    if response.status_code != 201:
        context.context(f"PUT source returned {response.status_code}, expected 201")
        return WebdavTckTestResult.FAILHARD

    # Copy it
    response = await session.client.copy(
        src, dest, overwrite=False, depth="infinity", test_name="copy_simple"
    )

    if response.status_code not in (200, 201, 204):
        context.context(
            f"COPY failed with {response.status_code}: expected 200, 201, or 204"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 201:
        context.warning("COPY to new resource should give 201 (RFC4918:S9.8.5)")

    # Verify destination exists with GET
    response = await session.client.get(dest, test_name="copy_simple_verify")
    if response.status_code != 200:
        context.context(f"GET copied resource returned {response.status_code}")
        return WebdavTckTestResult.FAIL

    if response.content != content:
        context.context("Copied content doesn't match source")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_copy_overwrite(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """COPY: Test overwrite behavior with Overwrite header."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    src = session.make_path("copysrc")
    dest = session.make_path("copydest")

    # Try to copy again with Overwrite: F (should fail)
    response = await session.client.copy(
        src, dest, overwrite=False, depth="infinity", test_name="copy_overwrite_false"
    )

    if response.status_code not in (400, 412):
        context.context(
            f"COPY with Overwrite: F on existing resource should fail with 412, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # Now with Overwrite: T (should succeed)
    response = await session.client.copy(
        src, dest, overwrite=True, depth="infinity", test_name="copy_overwrite_true"
    )

    if response.status_code not in (200, 204):
        context.context(
            f"COPY with Overwrite: T returned {response.status_code}, expected 204"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 204:
        context.warning(
            f"COPY to existing resource should give 204 (RFC4918:S9.8.5), got {response.status_code}"
        )

    return WebdavTckTestResult.OK


async def test_copy_nodestcoll(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """COPY: Copy into non-existent collection (expect 409)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    src = session.make_path("copysrc")
    nodest = session.make_path("nonesuch", "foo")

    response = await session.client.copy(
        src, nodest, overwrite=False, depth="0", test_name="copy_nodestcoll"
    )

    if response.status_code == 201:
        context.context(
            "COPY into non-existent collection succeeded (should have failed)"
        )
        return WebdavTckTestResult.FAIL

    if response.status_code != 409:
        context.warning(
            f"COPY to non-existent collection gave {response.status_code}, expected 409 (RFC4918:S9.8.5)"
        )

    return WebdavTckTestResult.OK


async def test_copy_coll(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """COPY: Recursive collection copy with Depth: infinity."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    csrc = session.make_path("ccsrc")
    cdest = session.make_path("ccdest")
    subsrc = session.make_path("ccsrc", "subcoll")

    # Create source collection
    response = await session.client.mkcol(csrc, test_name="copy_coll_setup")
    if response.status_code != 201:
        context.context(f"MKCOL for source collection failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Add multiple resources to the collection
    content = b"Resource content\n"
    for n in range(10):
        path = session.make_path("ccsrc", f"foo.{n}")
        response = await session.client.put(
            path, content, test_name=f"copy_coll_setup_{n}"
        )
        if response.status_code != 201:
            context.context(f"PUT resource {n} failed: {response.status_code}")
            return WebdavTckTestResult.FAILHARD

    # Create subcollection
    response = await session.client.mkcol(subsrc, test_name="copy_coll_setup_sub")
    if response.status_code != 201:
        context.context(f"MKCOL for subcollection failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Clean up destination if it exists
    await session.client.delete(cdest, test_name="copy_coll_cleanup")

    # Copy the collection with Depth: infinity
    response = await session.client.copy(
        csrc, cdest, overwrite=False, depth="infinity", test_name="copy_coll"
    )

    if response.status_code not in (200, 201, 204):
        context.context(f"Collection COPY failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Delete source to be sure we're checking the copy
    await session.client.delete(csrc, test_name="copy_coll_del_src")

    # Verify all resources were copied
    for n in range(10):
        path = session.make_path("ccdest", f"foo.{n}")
        response = await session.client.delete(path, test_name=f"copy_coll_verify_{n}")
        if response.status_code not in (200, 204):
            context.context(f"Copied resource {n} not found at {path}")
            return WebdavTckTestResult.FAIL

    # Verify subcollection was copied
    subdest = session.make_path("ccdest", "subcoll")
    response = await session.client.delete(subdest, test_name="copy_coll_verify_sub")
    if response.status_code not in (200, 204):
        context.context("Copied subcollection not found")
        return WebdavTckTestResult.FAIL

    # Cleanup
    await session.client.delete(cdest, test_name="copy_coll_cleanup")

    return WebdavTckTestResult.OK


async def test_copy_shallow(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """COPY: Non-recursive collection copy with Depth: 0."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    csrc = session.make_path("ccsrc_shallow")
    cdest = session.make_path("ccdest_shallow")

    # Create source collection with one member
    response = await session.client.mkcol(csrc, test_name="copy_shallow_setup")
    if response.status_code != 201:
        context.context(f"MKCOL failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    content = b"Member content\n"
    member_path = session.make_path("ccsrc_shallow", "foo")
    response = await session.client.put(
        member_path, content, test_name="copy_shallow_setup_member"
    )
    if response.status_code != 201:
        context.context(f"PUT member failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Clean destination
    await session.client.delete(cdest, test_name="copy_shallow_cleanup")

    # Copy with Depth: 0 (should copy collection but not members)
    response = await session.client.copy(
        csrc, cdest, overwrite=False, depth="0", test_name="copy_shallow"
    )

    if response.status_code not in (200, 201, 204):
        context.context(f"Shallow COPY failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Delete source
    await session.client.delete(csrc, test_name="copy_shallow_del_src")

    # Verify member was NOT copied
    dest_member = session.make_path("ccdest_shallow", "foo")
    response = await session.client.delete(dest_member, test_name="copy_shallow_verify")

    if response.status_code != 404:
        context.context(
            f"Shallow COPY should not have copied member, but DELETE returned {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # Cleanup
    await session.client.delete(cdest, test_name="copy_shallow_cleanup")

    return WebdavTckTestResult.OK


async def test_move(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MOVE: Basic resource move operations."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    src = session.make_path("movesrc")
    src2 = session.make_path("movesrc2")
    dest = session.make_path("movedest")

    # Create source resources
    content = b"Move source content\n"
    response = await session.client.put(src, content, test_name="move_setup")
    if response.status_code != 201:
        context.context(f"PUT source failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    response = await session.client.put(src2, content, test_name="move_setup2")
    if response.status_code != 201:
        context.context(f"PUT source2 failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Move first resource
    response = await session.client.move(src, dest, overwrite=False, test_name="move")

    if response.status_code not in (200, 201, 204):
        context.context(f"MOVE failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    if response.status_code != 201:
        context.warning(
            f"MOVE to new resource should give 201, got {response.status_code}"
        )

    # Verify source no longer exists
    response = await session.client.get(src, test_name="move_verify_src")
    if response.status_code != 404:
        context.context("MOVE source still exists after move")
        return WebdavTckTestResult.FAIL

    # Try MOVE with Overwrite: F to existing destination (should fail)
    response = await session.client.move(
        src2, dest, overwrite=False, test_name="move_overwrite_false"
    )

    if response.status_code not in (400, 412):
        context.context(
            f"MOVE with Overwrite: F should fail with 412, got {response.status_code}"
        )
        return WebdavTckTestResult.FAIL

    # MOVE with Overwrite: T (should succeed)
    response = await session.client.move(
        src2, dest, overwrite=True, test_name="move_overwrite_true"
    )

    if response.status_code not in (200, 204):
        context.context(f"MOVE with Overwrite: T failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    if response.status_code != 204:
        context.warning(
            f"MOVE to existing resource should give 204, got {response.status_code}"
        )

    # Cleanup
    await session.client.delete(dest, test_name="move_cleanup")

    return WebdavTckTestResult.OK


async def test_move_coll(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MOVE: Move collection with members."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    msrc = session.make_path("mvsrc")
    mdest = session.make_path("mvdest")
    subsrc = session.make_path("mvsrc", "subcoll")

    # Create source collection
    response = await session.client.mkcol(msrc, test_name="move_coll_setup")
    if response.status_code != 201:
        context.context(f"MKCOL failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Add resources to collection
    content = b"Resource content\n"
    for n in range(10):
        path = session.make_path("mvsrc", f"foo.{n}")
        response = await session.client.put(
            path, content, test_name=f"move_coll_setup_{n}"
        )
        if response.status_code != 201:
            context.context(f"PUT resource {n} failed: {response.status_code}")
            return WebdavTckTestResult.FAILHARD

    # Create subcollection
    response = await session.client.mkcol(subsrc, test_name="move_coll_setup_sub")
    if response.status_code != 201:
        context.context(f"MKCOL subcollection failed: {response.status_code}")
        return WebdavTckTestResult.FAILHARD

    # Move the collection
    response = await session.client.move(
        msrc, mdest, overwrite=False, test_name="move_coll"
    )

    if response.status_code not in (200, 201, 204):
        context.context(f"Collection MOVE failed: {response.status_code}")
        return WebdavTckTestResult.FAIL

    # Verify source no longer exists
    response = await session.client.get(msrc, test_name="move_coll_verify_src")
    if response.status_code != 404:
        context.context("MOVE source collection still exists")
        return WebdavTckTestResult.FAIL

    # Verify all members exist at destination
    for n in range(10):
        path = session.make_path("mvdest", f"foo.{n}")
        response = await session.client.delete(path, test_name=f"move_coll_verify_{n}")
        if response.status_code not in (200, 204):
            context.context(f"Moved resource {n} not found")
            return WebdavTckTestResult.FAIL

    # Verify subcollection was moved
    subdest = session.make_path("mvdest", "subcoll")
    response = await session.client.delete(subdest, test_name="move_coll_verify_sub")
    if response.status_code not in (200, 204):
        context.context("Moved subcollection not found")
        return WebdavTckTestResult.FAIL

    # Cleanup
    await session.client.delete(mdest, test_name="move_coll_cleanup")

    return WebdavTckTestResult.OK


def create_copymove_suite(session: WebdavTckSession) -> WebdavTckTestSuite:
    """Create the copymove test suite.

    Args:
        session: WebdavTck session

    Returns:
        Test suite with COPY/MOVE tests
    """
    suite = WebdavTckTestSuite("copymove", session=session)

    # Add tests in order
    tests = [
        test_copy_simple,
        test_copy_overwrite,
        test_copy_nodestcoll,
        test_copy_coll,
        test_copy_shallow,
        test_move,
        test_move_coll,
    ]

    for test_func in tests:
        suite.add_test(test_func)

    return suite



"""Basic WebDAV operations test suite.

Ports src/basic.c from the original litmus C implementation.
Tests fundamental HTTP and WebDAV methods for creating, retrieving,
and deleting resources and collections.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from webdav_tck.framework import (
    WebdavTckContext,
    WebdavTckTestResult,
    WebdavTckTestSuite,
)

if TYPE_CHECKING:
    from webdav_tck.session import WebdavTckSession


async def test_options(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """OPTIONS: Check OPTIONS request and DAV header."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    response = await session.client.options(session.base_path, test_name="options")

    # Accept 200 or 204 (both are valid per HTTP spec)
    if response.status_code not in (200, 204):
        context.context(f"OPTIONS returned {response.status_code}, expected 200 or 204")
        return WebdavTckTestResult.FAIL

    dav_header = response.headers.get("DAV", "")
    if not dav_header:
        context.context("No DAV header in OPTIONS response")
        return WebdavTckTestResult.FAIL

    if "1" not in dav_header:
        context.context(f"DAV header does not indicate Class 1 support: {dav_header}")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_put_get(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PUT/GET: Create resource with PUT, verify with GET."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("res1")
    content = b"This is a test file for litmus testing.\n"

    # PUT the resource
    response = await session.client.put(path, content, test_name="put_get")

    if response.status_code != 201:
        context.context(f"PUT returned {response.status_code}, expected 201")
        return WebdavTckTestResult.FAIL

    # GET the resource back
    response = await session.client.get(path, test_name="put_get")

    if response.status_code != 200:
        context.context(f"GET returned {response.status_code}, expected 200")
        return WebdavTckTestResult.FAIL

    # Verify content matches byte-for-byte
    if response.content != content:
        context.context("GET content does not match PUT content")
        return WebdavTckTestResult.FAIL

    # Store ETag if present
    etag = response.headers.get("ETag")
    if etag:
        session.store_etag(path, etag)

    return WebdavTckTestResult.OK


async def test_put_get_utf8_segment(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PUT/GET: UTF-8 character in URI (€ as %e2%82%ac)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Use UTF-8 character in path (€ = euro sign)
    path = session.make_path("res-%E2%82%AC")
    content = b"UTF-8 filename test\n"

    # PUT the resource
    response = await session.client.put(path, content, test_name="put_get_utf8_segment")

    if response.status_code != 201:
        context.context(f"PUT returned {response.status_code}, expected 201")
        return WebdavTckTestResult.FAIL

    # GET the resource back
    response = await session.client.get(path, test_name="put_get_utf8_segment")

    if response.status_code != 200:
        context.context(f"GET returned {response.status_code}, expected 200")
        return WebdavTckTestResult.FAIL

    # Verify content
    if response.content != content:
        context.context("GET content does not match PUT content")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_put_no_parent(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """PUT: PUT to non-existent parent collection (expect 409)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    # Try to PUT into a collection that doesn't exist
    path = session.make_path("noparent", "res")
    content = b"test\n"

    response = await session.client.put(path, content, test_name="put_no_parent")

    if response.status_code != 409:
        context.context(f"PUT returned {response.status_code}, expected 409 Conflict")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_mkcol(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MKCOL: Create collection."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("coll1")

    response = await session.client.mkcol(path, test_name="mkcol")

    if response.status_code != 201:
        context.context(f"MKCOL returned {response.status_code}, expected 201")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_mkcol_again(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MKCOL: Create same collection twice (should fail)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("coll1")

    # Try to create again (should fail)
    response = await session.client.mkcol(path, test_name="mkcol_again")

    if response.status_code in (200, 201):
        context.context(f"MKCOL returned {response.status_code}, should have failed")
        return WebdavTckTestResult.FAIL

    # Expected: 405 Method Not Allowed or similar
    if response.status_code not in (405, 409):
        context.warning(f"MKCOL returned {response.status_code}, expected 405 or 409")

    return WebdavTckTestResult.OK


async def test_delete(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """DELETE: Delete a resource."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("res1")

    response = await session.client.delete(path, test_name="delete")

    if response.status_code not in (200, 204):
        context.context(f"DELETE returned {response.status_code}, expected 200 or 204")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_delete_null(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """DELETE: Delete non-existent resource (expect 404)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("nothere")

    response = await session.client.delete(path, test_name="delete_null")

    if response.status_code != 404:
        context.context(f"DELETE returned {response.status_code}, expected 404")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_delete_coll(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """DELETE: Delete empty collection."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("coll1")

    response = await session.client.delete(path, test_name="delete_coll")

    if response.status_code not in (200, 204):
        context.context(f"DELETE returned {response.status_code}, expected 200 or 204")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_mkcol_no_parent(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MKCOL: Create collection in non-existent parent (expect 409)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("noparent", "coll")

    response = await session.client.mkcol(path, test_name="mkcol_no_parent")

    if response.status_code != 409:
        context.context(f"MKCOL returned {response.status_code}, expected 409 Conflict")
        return WebdavTckTestResult.FAIL

    return WebdavTckTestResult.OK


async def test_mkcol_with_body(
    session: WebdavTckSession, context: WebdavTckContext
) -> WebdavTckTestResult:
    """MKCOL: MKCOL with request body (expect 415)."""
    if not session.client:
        context.context("Client not initialized")
        return WebdavTckTestResult.FAILHARD

    path = session.make_path("coll2")
    body = b"<test>body</test>"

    response = await session.client.mkcol(path, test_name="mkcol_with_body", body=body)

    # Per RFC 4918, server must reject MKCOL with unsupported Content-Type
    if response.status_code == 201:
        context.context("MKCOL with body returned 201, should have failed")
        return WebdavTckTestResult.FAIL

    # Expected: 415 Unsupported Media Type
    if response.status_code != 415:
        context.warning(
            f"MKCOL with body returned {response.status_code}, expected 415"
        )

    return WebdavTckTestResult.OK


def create_basic_suite(session: WebdavTckSession) -> WebdavTckTestSuite:
    """Create the basic test suite.

    Args:
        session: WebdavTck session

    Returns:
        Test suite with basic tests
    """
    suite = WebdavTckTestSuite("basic", session=session)

    # Add tests in order
    tests = [
        test_options,
        test_put_get,
        test_put_get_utf8_segment,
        test_put_no_parent,
        test_mkcol,
        test_mkcol_again,
        test_delete,
        test_delete_null,
        test_delete_coll,
        test_mkcol_no_parent,
        test_mkcol_with_body,
    ]

    for test_func in tests:
        suite.add_test(test_func)

    return suite



"""WebDAV client wrapper around httpx."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

import httpx
from typing_extensions import Self

if TYPE_CHECKING:
    from types import TracebackType


class WebDAVResponse:
    """Wrapper for HTTP responses with WebDAV-specific helpers."""

    def __init__(self, response: httpx.Response):
        self._response = response

    @property
    def status_code(self) -> int:
        """Get HTTP status code."""
        return self._response.status_code

    @property
    def headers(self) -> httpx.Headers:
        """Get response headers."""
        return self._response.headers

    @property
    def content(self) -> bytes:
        """Get response body as bytes."""
        return self._response.content

    @property
    def text(self) -> str:
        """Get response body as text."""
        return self._response.text

    def raise_for_status(self) -> None:
        """Raise exception for 4xx/5xx status codes."""
        self._response.raise_for_status()

    def __repr__(self) -> str:
        return f"WebDAVResponse(status={self.status_code})"


class WebDAVClient:
    """WebDAV-aware HTTP client.

    Wraps httpx.AsyncClient with WebDAV-specific methods and headers.
    """

    def __init__(
        self,
        base_url: str,
        username: str | None = None,
        password: str | None = None,
        verify_ssl: bool = True,
        proxy: str | None = None,
        client_cert: str | None = None,
        litmus_header: str = "X-WebdavTck",
        timeout: float = 30.0,
    ):
        """Initialize WebDAV client.

        Args:
            base_url: Base URL for WebDAV server
            username: Optional username for authentication
            password: Optional password for authentication
            verify_ssl: Whether to verify SSL certificates
            proxy: Optional proxy URL
            client_cert: Optional path to client certificate
            litmus_header: Header name for test identification
            timeout: Request timeout in seconds
        """
        self.base_url = base_url.rstrip("/")
        self.litmus_header = litmus_header

        # Build auth
        auth = None
        if username and password:
            auth = httpx.BasicAuth(username, password)

        # Build client config
        client_kwargs: dict[str, Any] = {
            "auth": auth,
            "verify": verify_ssl,
            "timeout": timeout,
            "follow_redirects": False,
        }

        if proxy:
            client_kwargs["proxies"] = proxy

        if client_cert:
            client_kwargs["cert"] = client_cert

        self._client = httpx.AsyncClient(**client_kwargs)
        self._test_counter = 0

    async def __aenter__(self) -> Self:
        """Async context manager entry."""
        await self._client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Async context manager exit."""
        await self._client.__aexit__(exc_type, exc_value, traceback)

    async def close(self) -> None:
        """Close the client connection."""
        await self._client.aclose()

    def _make_url(self, path: str) -> str:
        """Make full URL from path.

        Args:
            path: Path relative to base URL

        Returns:
            Full URL
        """
        if path.startswith("http://") or path.startswith("https://"):
            return path
        return urljoin(self.base_url + "/", path.lstrip("/"))

    def _make_headers(self, test_name: str = "", **extra: str) -> dict[str, str]:
        """Make request headers with litmus identification.

        Args:
            test_name: Name of the test being run
            **extra: Additional headers

        Returns:
            Dictionary of headers
        """
        headers = {self.litmus_header: test_name or f"test-{self._test_counter}"}
        headers.update(extra)
        self._test_counter += 1
        return headers

    async def request(
        self,
        method: str,
        path: str,
        test_name: str = "",
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> WebDAVResponse:
        """Make an HTTP request.

        Args:
            method: HTTP method
            path: Path or full URL
            test_name: Name of test for header
            headers: Optional additional headers
            **kwargs: Additional arguments for httpx

        Returns:
            WebDAV response wrapper
        """
        url = self._make_url(path)
        req_headers = self._make_headers(test_name, **(headers or {}))

        response = await self._client.request(
            method, url, headers=req_headers, **kwargs
        )
        return WebDAVResponse(response)

    async def options(self, path: str, test_name: str = "") -> WebDAVResponse:
        """Send OPTIONS request.

        Args:
            path: Path to query
            test_name: Name of test

        Returns:
            Response with Allow and DAV headers
        """
        return await self.request("OPTIONS", path, test_name=test_name)

    async def get(
        self, path: str, test_name: str = "", headers: dict[str, str] | None = None
    ) -> WebDAVResponse:
        """Send GET request.

        Args:
            path: Path to resource
            test_name: Name of test
            headers: Optional additional headers

        Returns:
            Response with resource content
        """
        return await self.request("GET", path, test_name=test_name, headers=headers)

    async def put(
        self,
        path: str,
        content: bytes,
        test_name: str = "",
        headers: dict[str, str] | None = None,
        if_header: str | None = None,
    ) -> WebDAVResponse:
        """Send PUT request to create/update resource.

        Args:
            path: Path to resource
            content: Content to upload
            test_name: Name of test
            headers: Optional additional headers
            if_header: Optional If: header for conditional requests

        Returns:
            Response (typically 201 Created or 204 No Content)
        """
        all_headers = headers.copy() if headers else {}
        if if_header:
            all_headers["If"] = if_header
        return await self.request(
            "PUT", path, test_name=test_name, headers=all_headers, content=content
        )

    async def delete(self, path: str, test_name: str = "") -> WebDAVResponse:
        """Send DELETE request.

        Args:
            path: Path to resource
            test_name: Name of test

        Returns:
            Response (typically 204 No Content)
        """
        return await self.request("DELETE", path, test_name=test_name)

    async def mkcol(
        self, path: str, test_name: str = "", body: bytes | None = None
    ) -> WebDAVResponse:
        """Send MKCOL request to create a collection.

        Args:
            path: Path to new collection
            test_name: Name of test
            body: Optional request body (should typically be empty)

        Returns:
            Response (typically 201 Created)
        """
        return await self.request("MKCOL", path, test_name=test_name, content=body)

    async def copy(
        self,
        src: str,
        dest: str,
        overwrite: bool = True,
        depth: str = "infinity",
        test_name: str = "",
    ) -> WebDAVResponse:
        """Send COPY request.

        Args:
            src: Source path
            dest: Destination path (can be absolute or relative URL)
            overwrite: Whether to overwrite existing resource
            depth: Depth header value (0, 1, infinity)
            test_name: Name of test

        Returns:
            Response (typically 201 Created or 204 No Content)
        """
        headers = {
            "Destination": self._make_url(dest)
            if not dest.startswith("http")
            else dest,
            "Overwrite": "T" if overwrite else "F",
            "Depth": depth,
        }
        return await self.request("COPY", src, test_name=test_name, headers=headers)

    async def move(
        self,
        src: str,
        dest: str,
        overwrite: bool = True,
        test_name: str = "",
    ) -> WebDAVResponse:
        """Send MOVE request.

        Args:
            src: Source path
            dest: Destination path (can be absolute or relative URL)
            overwrite: Whether to overwrite existing resource
            test_name: Name of test

        Returns:
            Response (typically 201 Created or 204 No Content)
        """
        headers = {
            "Destination": self._make_url(dest)
            if not dest.startswith("http")
            else dest,
            "Overwrite": "T" if overwrite else "F",
        }
        return await self.request("MOVE", src, test_name=test_name, headers=headers)

    async def propfind(
        self,
        path: str,
        depth: int = 0,
        body: bytes | None = None,
        test_name: str = "",
    ) -> WebDAVResponse:
        """Send PROPFIND request.

        Args:
            path: Path to resource
            depth: Depth header value (0, 1, infinity)
            body: XML body with property names
            test_name: Name of test

        Returns:
            Response with multistatus XML
        """
        headers = {"Depth": str(depth)}
        return await self.request(
            "PROPFIND", path, test_name=test_name, headers=headers, content=body or b""
        )

    async def proppatch(
        self,
        path: str,
        body: bytes,
        test_name: str = "",
        if_header: str | None = None,
    ) -> WebDAVResponse:
        """Send PROPPATCH request.

        Args:
            path: Path to resource
            body: XML body with property updates
            test_name: Name of test
            if_header: Optional If: header for conditional requests

        Returns:
            Response with multistatus XML
        """
        headers = {"If": if_header} if if_header else None
        return await self.request(
            "PROPPATCH", path, test_name=test_name, headers=headers, content=body
        )

    async def lock(
        self,
        path: str,
        body: bytes | None,
        timeout: str = "Second-3600",
        depth: str = "0",
        if_header: str | None = None,
        test_name: str = "",
    ) -> WebDAVResponse:
        """Send LOCK request.

        Args:
            path: Path to resource
            body: XML body with lock info (None for lock refresh)
            timeout: Timeout header value
            depth: Depth header value
            if_header: Optional If header for lock refresh
            test_name: Name of test

        Returns:
            Response with lock token
        """
        headers = {"Timeout": timeout, "Depth": depth}
        if if_header:
            headers["If"] = if_header

        return await self.request(
            "LOCK", path, test_name=test_name, headers=headers, content=body
        )

    async def unlock(
        self,
        path: str,
        lock_token: str,
        test_name: str = "",
    ) -> WebDAVResponse:
        """Send UNLOCK request.

        Args:
            path: Path to resource
            lock_token: Lock token to release
            test_name: Name of test

        Returns:
            Response (typically 204 No Content)
        """
        headers = {"Lock-Token": f"<{lock_token}>"}
        return await self.request("UNLOCK", path, test_name=test_name, headers=headers)



"""Test framework for litmus."""

from __future__ import annotations

from webdav_tck.framework.result import WebdavTckTestResult
from webdav_tck.framework.runner import TestRunner
from webdav_tck.framework.suite import (
    WebdavTckContext,
    WebdavTckTestCase,
    WebdavTckTestSuite,
)

# For backwards compatibility, export as TestResult too
TestResult = WebdavTckTestResult

__all__ = [
    "TestResult",
    "TestRunner",
    "WebdavTckContext",
    "WebdavTckTestCase",
    "WebdavTckTestResult",
    "WebdavTckTestSuite",
]
